/*
 * Copyright ©2009 tpoinsot
 * Les sources qui constituent ce projet "Soma" de même que la documentation
 * associée sont la propriété de leur auteur.
 * Je donne mon accord au site developpez.com pour l'utilisation de tout ou
 * partie des sources et de la documentation de ce projet dans les pages developpez.com
 */

package soma;

import java.util.ArrayList;
import java.awt.Canvas;
import java.awt.Graphics;
import java.awt.Color;

/**
 *
 * @author tpoinsot
 */
public class CanvasSoma extends Canvas
{
    public ArrayList<Cube> forme = new ArrayList<Cube>();
    
    private double alpha = Math.PI/3; // != PI, != -PI/2

    public CanvasSoma()
    {
    }

    public void setForme(ArrayList<Cube> ql)
    {
        removeForme();
        for (int i = 0; i < ql.size(); i++)
            forme.add(ql.get(i));
        repaint();
    }

    public void removeForme()
    {
        forme.clear();
        repaint();
    }

    @Override
    public void paint(Graphics g)
    {
        int sz = 10; // nombre de points
        int a; // largeur des cubes élémentaires = largeur écran par unité

        if ( super.getWidth() < super.getHeight() ) {
            a = (int)(super.getWidth() / ( sz * (1 + Math.cos(alpha))) );
        }
        else {
            a = (int)(super.getHeight() / ( sz * (1 + Math.sin(alpha))) );
        }
        int gxorig = (int)(sz*a*Math.cos(alpha));
        int gyorig = sz*a;

        g.setColor(Color.white);
        g.fillRect(0, 0, super.getWidth(), super.getHeight());

        int acosalpha = (int)(a*Math.cos(alpha));
        int asinalpha = (int)(a*Math.sin(alpha));

        /*
         *   +----+
         *  /    /|
         * +----+ +
         * |    |/
         * +----+
         */
        g.setColor(Color.black);
        g.drawLine(gxorig, gyorig, gxorig+sz*a, gyorig);
        g.drawLine(gxorig, gyorig, gxorig-sz*acosalpha, gyorig+sz*asinalpha);
        g.drawLine(gxorig, gyorig, gxorig, 0);

        for (int i = 0; i < forme.size(); i++) {
            Cube q = forme.get(i);

            int gx = gxorig + q.x*a - q.z*acosalpha - acosalpha;
            int gy = gyorig - q.y*a + q.z*asinalpha + asinalpha;

            Color c, cb;
            switch ( q.numeroPiece ) {
                case 1: c=new Color(255,0,0); cb=new Color(128,0,0); break;
                case 2: c=new Color(255,255,0); cb=new Color(128,128,0); break;
                case 3: c=new Color(0,255,0); cb=new Color(0,128,0); break;
                case 4: c=new Color(0,255,255); cb=new Color(0,128,128); break;
                case 5: c=new Color(0,0,255); cb=new Color(0,0,128); break;
                case 6: c=new Color(192,192,0); cb=new Color(64,64,0); break;
                case 7: c=new Color(255,192,255); cb=new Color(128,64,128); break;
                default: c=Color.LIGHT_GRAY; cb=Color.LIGHT_GRAY; break;
            }

            /*
             * +----+
             * |    |
             * +----+
             */
            int [] px1 = { gx, gx    , gx + a, gx + a };
            int [] py1 = { gy, gy - a, gy - a, gy     };
            g.setColor(c); 
            g.fillPolygon(px1, py1, 4);
            g.setColor(Color.black);
            g.drawPolygon(px1, py1, 4);

            /*
             *   +----+
             *  /    /
             * +----+
             */
            int [] px2 = { gx    , gx + acosalpha    , gx + a + acosalpha, gx + a };
            int [] py2 = { gy - a, gy - a - asinalpha, gy - a - asinalpha, gy - a };
            g.setColor(c); 
            g.fillPolygon(px2, py2, 4);
            g.setColor(Color.black);
            g.drawPolygon(px2, py2, 4);

            /*
             *        +
             *       /|
             *      + +
             *      |/
             *      +
             */
            int [] px3 = { gx + a, gx + a, gx + a + acosalpha, gx + a + acosalpha };
            int [] py3 = { gy    , gy - a, gy - a - asinalpha, gy - asinalpha     };
            g.setColor(cb); 
            g.fillPolygon(px3, py3, 4);
            g.setColor(Color.black);
            g.drawPolygon(px3, py3, 4);
        }
    }
}
