/*
 * Copyright ©2009 tpoinsot
 * Les sources qui constituent ce projet "Soma" de même que la documentation
 * associée sont la propriété de leur auteur.
 * Je donne mon accord au site developpez.com pour l'utilisation de tout ou
 * partie des sources et de la documentation de ce projet dans les pages developpez.com
 */

package soma;

/**
 * Classe Cube : comme une classe Point, qui gèrerait des coordonnées dans
 * l'espace, avec des propriétés particulières :
 * - poids : pour trier
 * - numeroPiece : numéro de la pièce quand le Cube en est un constituant.
 *
 * @author tpoinsot
 */
public class Cube implements Comparable
{
    public int x, y, z;
    public int poids;
    public int numeroPiece;

    public Cube()
    {
    }

    public Cube(int x, int y, int z)
    {
        this.x = x;
        this.y = y;
        this.z = z;
    }

    public Cube(Cube q)
    {
        this.x = q.x;
        this.y = q.y;
        this.z = q.z;
    }

    /**
     * Calcule le carré de la norme
     * @return la valeur
     */
    public int norme2()
    {
        return x*x + y*y + z*z;
    }

    public int compareTo(Object o)
    {
        Cube p = (Cube)o;
        if ( this.poids == p.poids )
            return this.norme2() - p.norme2();

        return this.poids - p.poids;
    }

    /**
     * Crée un cube en multipliant une matrice 3x3 par le cube.
     * @param mat matrice à multiplier, dont les lignes sont mises bout à bout pour former un tableau de 9 entiers
     * @return le nouveau cube résultant de la multiplication
     */
    public Cube multMat(int [] mat) // 9 entiers !!!
    {
        return new Cube(
            mat[0]*x + mat[1]*y + mat[2]*z,
            mat[3]*x + mat[4]*y + mat[5]*z,
            mat[6]*x + mat[7]*y + mat[8]*z
                );
    }
    /**
     * Crée un cube en déplaçant le cube des coordonnées de q
     * @param q
     * @return le cube déplacé
     */
    public Cube move(Cube q)
    {
        return new Cube(x + q.x, y + q.y, z + q.z);
    }

    /**
     * Crée un cube en déplaçant le cube des coordonnées de -q
     * @param q
     * @return le cube déplacé
     */
    public Cube moveNeg(Cube q)
    {
        return new Cube(x - q.x, y - q.y, z - q.z);
    }

   /**
     * Indique si le cube est à côté de q sur l'axe des x, de y ou de z
     * @param q un cube
     * @return vrai si le cube est adjacent à q
     */
    public boolean estAdjacent(Cube q)
    {
        return (   ( (x == q.x + 1 || x == q.x - 1) && y == q.y && z == q.z )
                || ( x == q.x && (y == q.y + 1 || y == q.y - 1) && z == q.z )
                || ( x == q.x && y == q.y && (z == q.z + 1 || z == q.z - 1) ) );
    }
    
    /**
     * Vérifie si les coordonnées du cube sont les mêmes que celle du cube passé en paramètre
     * @param q cube à comparer
     * @return vrai si les coordonnées sont bien les mêmes que celles de q
     */
    public boolean positionIdentique(Cube q)
    {
        return (x == q.x && y == q.y && z == q.z);
    }

    @Override
    public String toString()
    {
        return "(" + x + ", " + y + ", " + z + ") Pièce = " + numeroPiece;
    }
}
