/*
 * Soma.java Classe principale
 *
 * Copyright ©2009 tpoinsot
 * Les sources qui constituent ce projet "Soma" de même que la documentation
 * associée sont la propriété de leur auteur.
 * Je donne mon accord au site developpez.com pour l'utilisation de tout ou
 * partie des sources et de la documentation de ce projet dans les pages developpez.com
 *
 * Created on 5 oct. 2009, 17:44:50
 *
 * Méthode : j'essaie toutes les combinaisons et j'arrête
 * dès que j'en trouve une correcte.
 * je prends la liste des points (27 au départ) et j'essaie chaque piece
 * sur la première position.
 * pour essayer une pièce, je la regarde suivant les 3 axes dans un sens puis
 * dans l'autre, et je la fais tourner 4 fois de PI/2 (nous n'avons que des
 * cubes, commes des dés, même assemblés en pièce plus complexe)
 * 3 axes x 2 sens x 4 positions = 24 positions possibles pour chaque piece.
 * Les pièces ont 3 ou 4 cubes, soit 72 ou 96 possibilités par pièce.
 * De manière récursive, je teste une pièce au premier point libre,
 * et si j'arrive à la placer je recommence avec un nouveau jeu de point libre
 * et de pièces à placer plus restreint : si tout est pris, j'ai gagné.
 * si aucune pièce ne convient, je sors de l'itération.
 *
 * Améliorations :
 * - les pieces symétriques comportent des "positions" identiques : j'élimine les doublons
 * - plutôt que de gérer un flag "utilisé" je retire les points de la liste "à pourvoir"
 * - idem avec les piéces
 * - les points libres sont triés par poids, calculés par le nombre de points
 * adjacents (distance de 1); un poids de 0 revient à une case isolée, ou devenue
 * isolée. Je détecte rapidement ce cas particulier, et abrege la récursion.
 * - les pièces de base sont essayées dans l'ordre de complexité décroissante
 * d'abord les pièces 3D, P en dernier car a plus de symétrie (A, B, P), puis
 * les 4-cubes Z, T, L, et terminer par le 3-cubes V : A, B, P - Z, T, L - V
 */

package soma;

import java.io.File;
import java.io.FileNotFoundException;
import java.util.Scanner;
import java.util.regex.*;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Date;
import javax.swing.*;

/**
 *
 * @author tpoinsot
 */
public class Soma extends javax.swing.JFrame {
    // pièces de base
    private ArrayList<Piece> pieces = new ArrayList<Piece>();

    private boolean calculerNbSol;

    /** Creates new form Soma */
    public Soma() {
        try {
            UIManager.setLookAndFeel(UIManager.getSystemLookAndFeelClassName());
        }
        catch (Exception e) {}

        initComponents();

        jLabel2.setText("");

        // Numérotation : dans l'ordre de présentation du cahier des charges
        // V:1, L:2, T:3, Z:4, A:5, B:6, P:7
        // les pièces seront utilisées dans l'ordre de l'array : je
        // trie "à la main" pour mettre le 3cubes en dernier et commencer par
        // les pièces avec le moins de symétrie

        Piece p4 = new Piece(5); // A
        p4.addCube(0, 0, 0);
        p4.addCube(0, 1, 0);
        p4.addCube(1, 0, 0);
        p4.addCube(1, 0, 1);
        pieces.add(p4);

        Piece p5 = new Piece(6); // B
        p5.addCube(0, 0, 0);
        p5.addCube(1, 0, 0);
        p5.addCube(0, 1, 0);
        p5.addCube(0, 1, 1);
        pieces.add(p5);

        Piece p6 = new Piece(7); // P
        p6.addCube(0, 0, 0);
        p6.addCube(1, 0, 0);
        p6.addCube(0, 1, 0);
        p6.addCube(0, 0, 1);
        pieces.add(p6);

        Piece p3 = new Piece(4); // Z
        p3.addCube(0, 0, 0);
        p3.addCube(1, 0, 0);
        p3.addCube(1, 1, 0);
        p3.addCube(2, 1, 0);
        pieces.add(p3);

        Piece p2 = new Piece(3); // T
        p2.addCube(0, 0, 0);
        p2.addCube(1, 0, 0);
        p2.addCube(1, 1, 0);
        p2.addCube(2, 0, 0);
        pieces.add(p2);

        Piece p1 = new Piece(2); // L
        p1.addCube(0, 0, 0);
        p1.addCube(1, 0, 0);
        p1.addCube(0, 1, 0);
        p1.addCube(2, 0, 0);
        pieces.add(p1);

        Piece p0 = new Piece(1); // V
        p0.addCube(0, 0, 0);
        p0.addCube(1, 0, 0);
        p0.addCube(0, 1, 0);
        pieces.add(p0);

        for (int i = 0; i < pieces.size(); i++) {
            pieces.get(i).préparer();
        }

        jTextField1.setText("");
        // pour exécution en test :
        //jTextField1.setText("formes\\SOMA_exemple.txt");
        //résoudre();
    }

    /** This method is called from within the constructor to
     * initialize the form.
     * WARNING: Do NOT modify this code. The content of this method is
     * always regenerated by the Form Editor.
     */
    @SuppressWarnings("unchecked")
    // <editor-fold defaultstate="collapsed" desc="Generated Code">//GEN-BEGIN:initComponents
    private void initComponents() {

        canvas1 = new CanvasSoma();
        jScrollPane1 = new javax.swing.JScrollPane();
        jTextArea1 = new javax.swing.JTextArea();
        jTextField1 = new javax.swing.JTextField();
        jLabel1 = new javax.swing.JLabel();
        jButton1 = new javax.swing.JButton();
        jLabel2 = new javax.swing.JLabel();
        jButton2 = new javax.swing.JButton();
        jCheckBox1 = new javax.swing.JCheckBox();
        jMenuBar1 = new javax.swing.JMenuBar();

        setDefaultCloseOperation(javax.swing.WindowConstants.EXIT_ON_CLOSE);
        setTitle("SOMA");
        setResizable(false);

        jTextArea1.setColumns(20);
        jTextArea1.setEditable(false);
        jTextArea1.setFont(new java.awt.Font("Tahoma", 0, 11));
        jTextArea1.setRows(27);
        jTextArea1.setBorder(null);
        jTextArea1.setMinimumSize(new java.awt.Dimension(3, 18));
        jScrollPane1.setViewportView(jTextArea1);

        jTextField1.setText("jTextField1");

        jLabel1.setText("Nom du fichier :");

        jButton1.setText("Parcourir");
        jButton1.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                jButton1ActionPerformed(evt);
            }
        });

        jButton2.setText("Résoudre");
        jButton2.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                jButton2ActionPerformed(evt);
            }
        });

        jCheckBox1.setBackground(getBackground());
        jCheckBox1.setText("Calculer le nombre de solutions");
        jCheckBox1.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                jCheckBox1ActionPerformed(evt);
            }
        });
        setJMenuBar(jMenuBar1);

        javax.swing.GroupLayout layout = new javax.swing.GroupLayout(getContentPane());
        getContentPane().setLayout(layout);
        layout.setHorizontalGroup(
            layout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
            .addGroup(layout.createSequentialGroup()
                .addContainerGap()
                .addGroup(layout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
                    .addComponent(jLabel2, javax.swing.GroupLayout.DEFAULT_SIZE, 718, Short.MAX_VALUE)
                    .addGroup(layout.createSequentialGroup()
                        .addComponent(jScrollPane1, javax.swing.GroupLayout.PREFERRED_SIZE, javax.swing.GroupLayout.DEFAULT_SIZE, javax.swing.GroupLayout.PREFERRED_SIZE)
                        .addPreferredGap(javax.swing.LayoutStyle.ComponentPlacement.RELATED)
                        .addComponent(canvas1, javax.swing.GroupLayout.DEFAULT_SIZE, 546, Short.MAX_VALUE))
                    .addGroup(layout.createSequentialGroup()
                        .addComponent(jLabel1)
                        .addPreferredGap(javax.swing.LayoutStyle.ComponentPlacement.RELATED)
                        .addGroup(layout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
                            .addComponent(jTextField1, javax.swing.GroupLayout.PREFERRED_SIZE, 554, javax.swing.GroupLayout.PREFERRED_SIZE)
                            .addComponent(jCheckBox1))
                        .addPreferredGap(javax.swing.LayoutStyle.ComponentPlacement.RELATED, javax.swing.GroupLayout.DEFAULT_SIZE, Short.MAX_VALUE)
                        .addGroup(layout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING, false)
                            .addComponent(jButton1, javax.swing.GroupLayout.Alignment.TRAILING, javax.swing.GroupLayout.DEFAULT_SIZE, javax.swing.GroupLayout.DEFAULT_SIZE, Short.MAX_VALUE)
                            .addComponent(jButton2, javax.swing.GroupLayout.Alignment.TRAILING, javax.swing.GroupLayout.DEFAULT_SIZE, javax.swing.GroupLayout.DEFAULT_SIZE, Short.MAX_VALUE))))
                .addContainerGap())
        );
        layout.setVerticalGroup(
            layout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
            .addGroup(javax.swing.GroupLayout.Alignment.TRAILING, layout.createSequentialGroup()
                .addContainerGap()
                .addGroup(layout.createParallelGroup(javax.swing.GroupLayout.Alignment.BASELINE)
                    .addComponent(jLabel1)
                    .addComponent(jButton1)
                    .addComponent(jTextField1, javax.swing.GroupLayout.PREFERRED_SIZE, javax.swing.GroupLayout.DEFAULT_SIZE, javax.swing.GroupLayout.PREFERRED_SIZE))
                .addPreferredGap(javax.swing.LayoutStyle.ComponentPlacement.RELATED)
                .addGroup(layout.createParallelGroup(javax.swing.GroupLayout.Alignment.BASELINE)
                    .addComponent(jButton2)
                    .addComponent(jCheckBox1))
                .addPreferredGap(javax.swing.LayoutStyle.ComponentPlacement.RELATED)
                .addComponent(jLabel2, javax.swing.GroupLayout.PREFERRED_SIZE, 20, javax.swing.GroupLayout.PREFERRED_SIZE)
                .addPreferredGap(javax.swing.LayoutStyle.ComponentPlacement.RELATED)
                .addGroup(layout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
                    .addComponent(canvas1, javax.swing.GroupLayout.DEFAULT_SIZE, 482, Short.MAX_VALUE)
                    .addComponent(jScrollPane1, javax.swing.GroupLayout.DEFAULT_SIZE, 482, Short.MAX_VALUE))
                .addContainerGap())
        );

        pack();
    }// </editor-fold>//GEN-END:initComponents

    private void jButton1ActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_jButton1ActionPerformed
        try {
            File f0 = new File(jTextField1.getText());
            JFileChooser fc = new JFileChooser(f0.getPath());
            fc.setFileSelectionMode(JFileChooser.FILES_ONLY);
            int returnVal = fc.showOpenDialog(Soma.this);

            if ( returnVal == JFileChooser.APPROVE_OPTION ) {
                jTextField1.setText(fc.getSelectedFile().getPath());
                résoudre();
            }
        }
        catch (Exception e) {
            jLabel2.setText("Erreur : "+e.getMessage());
        }
    }//GEN-LAST:event_jButton1ActionPerformed

    private void jButton2ActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_jButton2ActionPerformed
        résoudre();
    }//GEN-LAST:event_jButton2ActionPerformed

    private void jCheckBox1ActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_jCheckBox1ActionPerformed
        // TODO add your handling code here:
        calculerNbSol = !calculerNbSol;
    }//GEN-LAST:event_jCheckBox1ActionPerformed

    ArrayList<Cube> lireForme(String nomFichier)
    {
        // lecture du fichier pour constituer la liste des points

        try {
            int nLigne = 0;
            int x, y, z;

            ArrayList<Cube> forme = new ArrayList<Cube>();
            Scanner sc = new Scanner(new File(nomFichier));

            for (nLigne = 1; sc.hasNextLine(); nLigne++ ) {
                sc.findInLine("(\\d+),(\\d+),(\\d+)");
                MatchResult result = sc.match();
                if ( result.groupCount() == 3 ) {
                    try {
                        x = Integer.parseInt( result.group(1) );
                        y = Integer.parseInt( result.group(2) );
                        z = Integer.parseInt( result.group(3) );
                        if ( x < 0 || x > 10 || y < 0 || y > 10  || z < 0 || z > 10 ) {
                            jLabel2.setText("Coordonnées refusées ligne "+nLigne);
                            return null;
                        }
                        forme.add(new Cube( x, y, z ));
                    }
                    catch (NumberFormatException en ) {
                        jLabel2.setText("Format numérique incorrect ligne "+nLigne);
                        return null;
                    }
                }
                if ( sc.hasNextLine() )
                    sc.nextLine();
            }
            sc.close();
            if ( forme.size() != 27 ) {
                jLabel2.setText("Pas de traitement : le fichier comporte "+forme.size()+" point(s) au lieu de 27.");
                return null;
            }
            return forme;
        }
        catch ( IllegalStateException e ) {
            jLabel2.setText("Le fichier n'est pas conforme ("+e.getLocalizedMessage()+")");
        }
        catch (FileNotFoundException e) {
            jLabel2.setText("Le fichier n'a pas été trouvé");
        }
        catch (Exception e) {
            jLabel2.setText("Erreur : "+e.getMessage());
        }
        return null;
    }

    @SuppressWarnings("unchecked")
    void résoudre()
    {
        try {
            Date dtDepart = new Date(); //============== Début chrono.

            jLabel2.setText(null);
            jTextArea1.setText(null);
            ((CanvasSoma)canvas1).removeForme();

            ArrayList<Cube> forme = lireForme(jTextField1.getText());

            if ( forme != null ) {
                Date dtInter = new Date(); //============== temps intermédiaire
                String msg;

                if ( poserPieces(forme, pieces, false) == 0 ) {
                    msg = "La résolution est impossible";
                }
                else {
                    if ( calculerNbSol ) {
                        ArrayList<Cube> forme2 = new ArrayList<Cube>(forme.size());
                        for (int i = 0; i < forme.size(); i++) {
                            forme2.add(new Cube(forme.get(i)));
                        }

                        // calcul du nombre de solutions
                        long nbSol = poserPieces(forme2, pieces, true);
                        msg = "OUI : voici la première ("+(dtInter.getTime()-dtDepart.getTime())+" ms) solution trouvée de " + nbSol;
                    }
                    else {
                        msg = "OUI : voici la première solution trouvée";
                    }
                }
                
                Date dtFin = new Date(); //============== Fin chrono.

                jLabel2.setText(msg+" ("+(dtFin.getTime()-dtDepart.getTime())+" ms)");

                // Afficher la liste des points avec leurs coordonnées
                for (int i = 0; i < forme.size(); i++) {
                    jTextArea1.append(i+" "+forme.get(i).toString()+"\n");
                }

                // Afficher le dessin de la forme
                for (int i = 0; i < forme.size(); i++)
                    forme.get(i).poids = 0;
                Collections.sort(forme);
                ((CanvasSoma)canvas1).setForme(forme);

           }
        }
        catch (Exception e) {
            jLabel2.setText("Erreur : "+e.getMessage());
        }
    }

    @SuppressWarnings("unchecked")
    long poserPieces(ArrayList<Cube> cubesLibres, ArrayList<Piece> piecesDisponibles, boolean comptage)
    {
        if ( cubesLibres == null || piecesDisponibles == null )
            return 0;
        if ( cubesLibres.size() == 0 || piecesDisponibles.size() == 0 )
            return 1;

        int i, j;

        // Calcul des poids = nombre de cases adjacentes (sur les axes) libre,
        // de 0 à 6.
        // exemple : 2 points côte à côte sur x ont chacun un poids de 1
        // 0 = aucune case libre à proximité, cas d'erreur.
        int nql = cubesLibres.size();
        for (i = 0; i < nql; i++)
            cubesLibres.get(i).poids = 0;

        for (i = 0; i < nql - 1; i++) {
            Cube pti = cubesLibres.get(i);

            for (j = i + 1; j < nql; j++) {
                Cube ptj = cubesLibres.get(j);
                if ( ptj.estAdjacent(pti) == true ) {
                    pti.poids++;
                    ptj.poids++;
                }
            }
        }

        // tri par poids croissant, et, à poids égal, par distance à l'origine (suppose que toutes
        // les coordonnées sont >= 0, ce qui est prévu !
        // + loin, je vais tenter de placer les pièces sur ce 1er élément de poids le + faible.
        Collections.sort(cubesLibres);
        // si le + petit poids est 0 : case isolée, donc impossible de résoudre
        if ( cubesLibres.get(0).poids == 0 )
            return 0;

        int ni = piecesDisponibles.size();
        long solutions = 0;
        for (i = 0; i < ni; i++)
        {
            Piece pi = piecesDisponibles.remove(i);
            
            // j'essaye chaque position de pi à la première case libre
            for (Piece p = pi.positionFirst(); p != null; p = pi.positionNext() )
            {
                ArrayList<Cube> CubesRestant = p.placer(cubesLibres);
                if ( CubesRestant != null )
                {
                    // la pièce est placée, cubesLibres s'est réduite en CubesRestant
                    // les poids vont changer mais pas ma liste cubesLibres
                    // la pièce pi est retirée de piecesRestantes, donc pi.position
                    // reste inchangé.
                    long nbSol = poserPieces(CubesRestant, piecesDisponibles, comptage);
                    if ( nbSol > 0 ) {
                        // toutes les pièces sont placées
                        if ( comptage == false ) {
                            piecesDisponibles.add(i, pi); // on replace pi là où on l'a prise
                            return 1;
                        }
                        else {
                            solutions += nbSol;
                        }
                    }
                    // sinon, la pièce rentre mais ne permet pas le placement
                    // des pièces restantes. Je nettoie et continue
                    p.libérer(cubesLibres);
                }
            }
            piecesDisponibles.add(i, pi); // on replace pi là où on l'a prise
        }
        return solutions;
    }

    /**
    * @param args the command line arguments
    */
    public static void main(String args[]) {
        java.awt.EventQueue.invokeLater(new Runnable() {
            public void run() {
                new Soma().setVisible(true);
            }
        });
    }

    // Variables declaration - do not modify//GEN-BEGIN:variables
    private java.awt.Canvas canvas1;
    private javax.swing.JButton jButton1;
    private javax.swing.JButton jButton2;
    private javax.swing.JCheckBox jCheckBox1;
    private javax.swing.JLabel jLabel1;
    private javax.swing.JLabel jLabel2;
    private javax.swing.JMenuBar jMenuBar1;
    private javax.swing.JScrollPane jScrollPane1;
    private javax.swing.JTextArea jTextArea1;
    private javax.swing.JTextField jTextField1;
    // End of variables declaration//GEN-END:variables

}
